@@ -20,7 +20,6 @@
 */
 
 #include "includes.h"
-#include "lib/events/events.h"
 #include "system/filesys.h"
 #include "popt.h"
 #include "cmdline.h"
@@ -50,7 +49,7 @@ static unsigned int pnn;
 
 static TDB_DATA old_data;
 
-static int success = true;
+static bool success = false;
 
 static void print_counters(void)
 {
@@ -79,6 +78,7 @@ static void check_counters(struct ctdb_context *ctdb, TDB_DATA data)
 {
 	int i;
 	uint32_t *counters, *old_counters;
+	bool monotonous = true;
 
 	counters     = (uint32_t *)data.dptr;
 	old_counters = (uint32_t *)old_data.dptr;
@@ -88,7 +88,7 @@ static void check_counters(struct ctdb_context *ctdb, TDB_DATA data)
 		if (counters[i]<old_counters[i]) {
 			printf("[%4u] ERROR: counters has decreased for node %u  From %u to %u\n", 
 			       getpid(), i, old_counters[i], counters[i]);
-			success = false;
+			monotonous = false;
 		}
 	}
 
@@ -99,6 +99,8 @@ static void check_counters(struct ctdb_context *ctdb, TDB_DATA data)
 
 	memcpy(old_data.dptr, data.dptr, data.dsize);
 	if (verbose) print_counters();
+
+	success = monotonous;
 }
 
 
@@ -135,7 +137,7 @@ static void test_store_records(struct ctdb_context *ctdb, struct event_context *
 			DEBUG(DEBUG_ERR, ("Failed to start transaction on node %d\n",
 			       ctdb_get_pnn(ctdb)));
 			talloc_free(tmp_ctx);
-			return;
+			continue;
 		}
 		if (verbose) DEBUG(DEBUG_ERR, ("transaction started\n"));
 		do_sleep(delay);
@@ -249,7 +251,7 @@ int main(int argc, const char *argv[])
 
 	ev = event_context_init(NULL);
 
-	ctdb = ctdb_cmdline_client(ev);
+	ctdb = ctdb_cmdline_client(ev, timeval_current_ofs(3, 0));
 	if (ctdb == NULL) {
 		DEBUG(DEBUG_ERR, ("Could not attach to daemon\n"));
 		return 1;
@@ -257,9 +259,11 @@ int main(int argc, const char *argv[])
 
 	/* attach to a specific database */
 	if (unsafe_writes == 1) {
-		ctdb_db = ctdb_attach(ctdb, "transaction.tdb", true, TDB_NOSYNC);
+		ctdb_db = ctdb_attach(ctdb, timeval_current_ofs(2, 0),
+				      "transaction.tdb", true, TDB_NOSYNC);
 	} else {
-		ctdb_db = ctdb_attach(ctdb, "transaction.tdb", true, 0);
+		ctdb_db = ctdb_attach(ctdb, timeval_current_ofs(2, 0),
+				      "transaction.tdb", true, 0);
 	}
 
 	if (!ctdb_db) {
